/*
**  LLVM if statement branching
**  Copyright (c) 2016 Magnus Norddahl
**
**  This software is provided 'as-is', without any express or implied
**  warranty.  In no event will the authors be held liable for any damages
**  arising from the use of this software.
**
**  Permission is granted to anyone to use this software for any purpose,
**  including commercial applications, and to alter it and redistribute it
**  freely, subject to the following restrictions:
**
**  1. The origin of this software must not be misrepresented; you must not
**     claim that you wrote the original software. If you use this software
**     in a product, an acknowledgment in the product documentation would be
**     appreciated but is not required.
**  2. Altered source versions must be plainly marked as such, and must not be
**     misrepresented as being the original software.
**  3. This notice may not be removed or altered from any source distribution.
**
*/

#pragma once

#include "ssa_bool.h"
#include "ssa_phi.h"

class SSAIfBlock
{
public:
	SSAIfBlock();
	void if_block(SSABool true_condition);
	void else_block();
	void end_block();

private:
	llvm::BasicBlock *if_basic_block;
	llvm::BasicBlock *else_basic_block;
	llvm::BasicBlock *end_basic_block;
};

template<typename T>
T ssa_min(T a, T b)
{
	SSAPhi<T> phi;
	SSAIfBlock if_block;
	if_block.if_block(a <= b);
	phi.add_incoming(a);
	if_block.else_block();
	phi.add_incoming(b);
	if_block.end_block();
	return phi.create();
}

template<typename T>
T ssa_max(T a, T b)
{
	SSAPhi<T> phi;
	SSAIfBlock if_block;
	if_block.if_block(a >= b);
	phi.add_incoming(a);
	if_block.else_block();
	phi.add_incoming(b);
	if_block.end_block();
	return phi.create();
}
